"""Utility functions to simplify life.
"""

from datetime import datetime, timezone
import json
import logging
import os
from pathlib import Path
import re

import fasteners

from seamm_dashboard import db, datastore, options
from seamm_datastore.database.models import Job

logger = logging.getLogger("__file__")


def get_job_id(filename):
    """Get the next job id from the given file.

    This uses the fasteners module to provide locking so that
    only one job at a time can access the file, so that the job
    ids are unique and monotonically increasing.
    """

    filename = os.path.expanduser(filename)

    lock_file = filename + ".lock"
    lock = fasteners.InterProcessLock(lock_file)
    locked = lock.acquire(blocking=True, timeout=5)

    if locked:
        if not os.path.isfile(filename):
            job_id = 1
            with open(filename, "w") as fd:
                fd.write("!MolSSI job_id 1.0\n")
                fd.write("1\n")
            lock.release()
        else:
            with open(filename, "r+") as fd:
                line = fd.readline()
                pos = fd.tell()
                if line == "":
                    lock.release()
                    raise EOFError("job_id file '{}' is empty".format(filename))
                line = line.strip()
                match = re.fullmatch(r"!MolSSI job_id ([0-9]+(?:\.[0-9]+)*)", line)
                if match is None:
                    lock.release()
                    raise RuntimeError(
                        "The job_id file has an incorrect header: {}".format(line)
                    )
                line = fd.readline()
                if line == "":
                    lock.release()
                    raise EOFError("job_id file '{}' is truncated".format(filename))
                try:
                    job_id = int(line)
                except TypeError:
                    raise TypeError(
                        "The job_id in file '{}' is not an integer: {}".format(
                            filename, line
                        )
                    )
                job_id += 1
                fd.seek(pos)
                fd.write("{:d}\n".format(job_id))
    else:
        raise RuntimeError("Could not lock the job_id file '{}'".format(filename))

    return job_id


def safe_filename(filename):
    # Make a single filename for a file transferred for a job
    if filename[0] == "~":
        path = Path(filename).expanduser()
    else:
        path = Path(filename)
    if path.anchor == "":
        result = "_".join(path.parts)
    else:
        result = "_".join(path.parts[1:])
    return "job:data/" + result


def setup_job(flowchart, project_names, title, description, parameters):
    """Add a new job to the queue.

    Parameters
    ----------
    flowchart : str
        The flowchart
    project_names : [str]
        The projects associated with this job.
    title : str
        The title of the job (<100 chars)
    description : str
        A longer description of the job.
    parameters : {str: ....}
        The parameters, including command line, for the job.

    Returns
    -------
    The job object
    """

    # Get the unique ID for the job...
    if options["job_id_file"] is None:
        job_id_file = os.path.join(datastore, "job.id")
    else:
        job_id_file = options["job_id_file"]
    job_id = get_job_id(job_id_file)

    # Create the directory and write the flowchart to it.
    project_path = Path(datastore).expanduser() / "projects" / project_names[0]
    directory = project_path / "Job_{:06d}".format(job_id)
    logger.info("Writing job files to " + str(directory))
    directory.mkdir(parents=True, exist_ok=True)

    path = directory / "flowchart.flow"
    with path.open("w") as fd:
        fd.write(flowchart)
    flowchart_file = str(path)

    # Write the json data file for the job
    data = {
        "data_version": "1.0",
        "command line": parameters["cmdline"],
        "title": title,
        "working directory": str(directory),
        "state": "submitted",
        "projects": project_names,
        "datastore": datastore,
        "job id": job_id,
        "submitted": datetime.now(timezone.utc).isoformat(),
    }
    path = directory / "job_data.json"
    with path.open("w") as fd:
        fd.write("!MolSSI job_data 1.0\n")
        json.dump(data, fd, sort_keys=True, indent=3)

    job = Job.create(
        job_id,
        path=str(directory),
        flowchart_filename=flowchart_file,
        project_names=project_names,
        title=title,
        description=description,
        parameters=parameters,
    )

    db.session.add(job)
    db.session.commit()

    return job



#!/usr/bin/env python
from seamm_dashboard import create_app, options
from waitress import serve


def run():
    app = create_app()

    # serve using waitress
    if "debug" in options:
        app.run(debug=True, use_reloader=True)  
    else:   
        # serve using waitress
        if options["localhost"]:
            serve(app, listen=f"localhost:{options['port']}", threads=12)
        else:
            serve(app, port=options["port"], threads=12)


if __name__ == "__main__":
    run()



import json
from pathlib import Path
import shlex

import requests

from flask import request, render_template, flash, redirect, url_for
from flask_jwt_extended import jwt_required, get_current_user
from flask_wtf import FlaskForm
from flask_wtf.file import FileField
from wtforms import (
    BooleanField,
    FloatField,
    IntegerField,
    MultipleFileField,
    SelectField,
    SelectMultipleField,
    StringField,
    SubmitField,
    TextAreaField,
)
from wtforms.validators import DataRequired, Length, Regexp

from seamm_datastore.database.models import Job, Project, UserProjectAssociation
from seamm_dashboard import db, authorize
from seamm_dashboard.routes.jobs.forms import EditJob, ImportJob, GetFlowchart
from seamm_dashboard.routes.api.auth import refresh_expiring_jwts
import seamm_dashboard.util as util
from . import jobs

flowchart_text = None
types = {
    "str": StringField,
    "int": IntegerField,
    "float": FloatField,
    "bool": BooleanField,
    "file": FileField,
}


@jobs.route("/views/jobs/")
@jobs.route("/views//jobs/")
@jwt_required(optional=True)
def jobs_list():
    return render_template("jobs/jobs_list.html", project=False)


@jobs.route("/views/jobs/<id>")
@jobs.route("/views//jobs/<id>")
@jobs.route("/views/jobs/<id>?filename=<filename>")
@jwt_required(optional=True)
def job_details(id, filename=None):

    from seamm_datastore.util import NotAuthorizedError

    try:
        job = Job.get_by_id(id)
    except NotAuthorizedError:
        return render_template("401.html")

    edit_job = authorize.update(job)

    own_string = "You do not own this job."

    if job.owner == get_current_user():
        own_string = "You are the job owner."

    # Build the url ourselves.
    base_url = url_for("main.index")
    edit_url = base_url + f"jobs/{id}/edit"
    submit_url = base_url + f"jobs/{id}/resubmit"

    # Figure out if we can use cdn for plotly
    plotly_location = "https://cdn.plot.ly/plotly-2.9.0.min.js"

    try:
        found_plotly = requests.get(f"{plotly_location}").status_code == 200
    except:
        # if the request doesn't work for some reason, just use the packaged plotly.
        found_plotly = False

    if not found_plotly:
        plotly_location = url_for(
            "static", filename="node_modules/plotly.js-dist-min/plotly.min.js"
        )

    return render_template(
        "jobs/job_report.html",
        edit_job=edit_job,
        job=job,
        edit_url=edit_url,
        submit_url=submit_url,
        own_string=own_string,
        plotly=plotly_location,
    )


@jobs.route("/jobs/<job_id>/edit", methods=["GET", "POST"])
@jwt_required(optional=True)
def edit_job(job_id):

    job = Job.query.get(job_id)

    if not authorize.update(job):
        return render_template("401.html")

    form = EditJob()

    # Build the url ourselves.
    base_url = url_for("main.index")
    job_url = base_url + f"#jobs/{job_id}"

    if form.validate_on_submit():
        job.title = form.name.data
        job.description = form.notes.data
        db.session.commit()
        flash("Job updated successfully.", "successs")

        return redirect(job_url)

    elif request.method == "GET":
        form.name.data = job.title
        form.notes.data = job.description

    return render_template(
        "jobs/edit_job.html", title=f"Edit Job {job_id}", form=form, back_url=job_url
    )


@jobs.route("/jobs/import", methods=["GET", "POST"])
@jwt_required()
def import_job():
    # TODO - rewrite
    form = ImportJob()

    if form.validate_on_submit():
        data = json.load(form.outfile.data.stream)

        working_directory = data["working directory"]

        job_info = Job.parse_job_data(working_directory)

        if form.title.data:
            job_info["title"] = form.title.data

        projects = data["projects"]

        project_objs = []
        for project in projects:
            query_project = Project.query.filter_by(name=project).one_or_none()

            # Create project if it doesn't exist
            if not query_project:
                query_project = Project(name=project)
                db.session.add(query_project)
                db.session.commit()

            project_objs.append(query_project)

        # TODO Add job
        # add_job(working_directory, job_info["title"], project_objs)
        flash(f"Job {job_info['id']} successfully added to database.")
        return redirect(url_for("main.index"))

    return render_template("jobs/import_job.html", form=form)


@jobs.route("/jobs/submit", methods=["GET", "POST"])
@jwt_required()
def submit_job():
    """Handle the dialogs for submitting a job by uploading a flowchart.
    
    Steps:
        1. Get the file
        2. Read the file to setup the dialog with the correct parameters
    """
    global flowchart_text
    form = GetFlowchart()

    if form.validate_on_submit():
        data = form.outfile.data.stream.read().decode()

        lines = data.splitlines()

        # Check the definition of the file
        if len(lines[0]) > 2 and lines[0][0:2] != "#!":
            header = lines[0]
        else:
            header = lines[1]
        fields = header.split()
        if len(fields) < 3 or fields[0] != "!MolSSI" or fields[1] != "flowchart":
            flash(f"The file is not a flowchart: {header}")
        else:
            flowchart_text = data
            return redirect(url_for("jobs.job_parameters"))

    return render_template("jobs/submit_job.html", form=form)


@jobs.route("/jobs/parameters", methods=["GET", "POST"])
@jwt_required()
def job_parameters(template_job=None):
    """Handle the dialogs for submitting a job by uploading a flowchart.

    This is step 2: setting up the page to get the parameters
    
    Steps:
        1. Get the file
        2. Read the file to setup the dialog with the correct parameters
    """
    global flowchart_text
    lines = flowchart_text.splitlines()

    # Find the definition of the file
    if len(lines[0]) > 2 and lines[0][0:2] != "#!":
        header = lines[0]
    else:
        header = lines[1]
    fields = header.split()
    if len(fields) < 3 or fields[0] != "!MolSSI" or fields[1] != "flowchart":
        flash(f"The file is not a flowchart: {header}")
        form = GetFlowchart()
        return render_template("jobs/submit_job.html", form=form)
    else:
        sections = {}
        section = ""
        saved = []
        for line in lines:
            if len(line) > 1 and line[0] == "#":
                if section != "":
                    sections[section] = "\n".join(saved)
                saved = []
                section = line[1:].split()[0]
            elif section != "":
                saved.append(line)
        if section != "" and len(saved) > 0:
            sections[section] = "\n".join(saved)
        if "flowchart" not in sections:
            flash("There was no flowchart in the file!")
            form = GetFlowchart()
            return render_template("jobs/submit_job.html", form=form)
        else:
            # Get info on the user and allowed projects, etc.
            projects = Project.get(
                permission="update",
                description=None,
                offset=None,
                limit=None,
                sort_by="name",
                order="asc",
            )
            project_names = sorted([x.name for x in projects])

            # Create an empty Form class
            class JP(FlaskForm):
                pass

            flowchart = json.loads(sections["flowchart"])

            # Find the control parameters, if any in the flowchart.
            # TODO: This doesn't check if the step is connected

            widgets = {}
            for step in flowchart["nodes"]:
                if step["class"] == "ControlParameters":
                    # Insert the correct input into the form
                    variables = step["attributes"]["parameters"]["variables"]["value"]
                    count = 0
                    for name, value in variables.items():
                        count += 1
                        _type = value["type"]
                        choices = value["choices"]
                        if isinstance(choices, str):
                            choices = json.loads(choices)
                        if len(choices) == 0:
                            choices = None

                        if _type == "str" and choices is not None:
                            if "or more" in value["nargs"]:
                                val = SelectMultipleField(name, choices=choices)
                            else:
                                val = SelectField(name, choices=choices)
                        elif _type == "file":
                            if "or more" in value["nargs"]:
                                val = MultipleFileField(name)
                            else:
                                val = FileField(name)
                        else:
                            val = types[_type](name)

                        widget = f"w{count}"
                        setattr(JP, widget, val)

                        widgets[widget] = {
                            "default": value["default"],
                            "description": value["help"],
                            "name": name,
                            **value,
                        }

            # And the final inputs for the job itself.
            setattr(JP, "project", SelectMultipleField("Project", choices=project_names))
            setattr(JP, "title", StringField("Job Title"))
            setattr(JP, "description", TextAreaField("Description"))
            setattr(JP, "submit", SubmitField("Submit Just This Job"))
            setattr(JP, "submit_more", SubmitField("Submit More Jobs..."))

            form = JP()

    if request.method == "GET":
        values = {}
        if template_job is not None:
            tmp = template_job.parameters
            if "control parameters" in tmp:
                values = tmp["control parameters"]

        for key, tmp in widgets.items():
            name = tmp["name"]
            if name in values:
                form[key].data = values[name]
            else:
                form[key].data = tmp["default"]

            if tmp["help"] != "":
                form[key].description = tmp["help"]

            if tmp["default"] != "":
                form[key].default = value["default"]

        if "metadata" in sections:
            data = json.loads(sections["metadata"])
            form.title.render_kw = {
                "size": "100",
                "maxlength": 100,
                "placeholder": "A short (< 100 chars) description of this job",
            }
            if template_job is not None and template_job.title != "":
                form.title.data = template_job.title
            elif data["title"] != "":
                form.title.data = data["title"]

            form.description.render_kw = {
                "rows": 8,
                "placeholder": "A longer description of this job",
            }
            if template_job is not None and template_job.description != "":
                form.description.data = template_job.description
            elif data["description"] != "":
                form.description.data = data["description"]
        if len(project_names) > 0:
            form.project.data = project_names[0]
    elif form.validate_on_submit():
        result = form.data

        # Prepare the command line arguments, transforming and remembering files
        files = {}
        control_parameters = {}
        if len(widgets) == 0:
            cmdline = []
        else:
            # Build the command line
            optional = []
            required = []
            for key, values in widgets.items():
                name = values["name"]
                if values["optional"] == "Yes":
                    if values["type"] == "bool":
                        control_parameters[name] = result[key]
                        if result[key] == "Yes":
                            optional.append(f"--{name}")
                    elif values["type"] == "file":
                        if values["nargs"] == "a single value":
                            file_storage = form[key].data
                            filename = file_storage.filename
                            if filename in files:
                                c = 2
                                tmp_name = f"{c}__{filename}"
                                while tmp_name in files:
                                    c += 1
                                    tmp_name = f"{c}__{filename}"
                                filename = tmp_name
                            files[filename] = file_storage
                            optional.append(f"--{name}")
                            optional.append(f"job:data/{filename}")
                        else:
                            first = True
                            for _file in form[key].data:
                                file_storage = _file
                                filename = file_storage.filename
                                if filename == "":
                                    continue
                                if filename in files:
                                    c = 2
                                    tmp_name = f"{c}__{filename}"
                                    while tmp_name in files:
                                        c += 1
                                        tmp_name = f"{c}__{filename}"
                                    filename = tmp_name
                                files[filename] = file_storage
                                if first:
                                    optional.append(f"--{name}")
                                    first = False
                                optional.append(f"job:data/{filename}")
                    else:
                        control_parameters[name] = result[key]
                        optional.append(f"--{name}")
                        if values["nargs"] == "a single value":
                            optional.append(result[key])
                        else:
                            optional.extend(shlex.split(result[key]))
                else:
                    if values["type"] == "file":
                        if values["nargs"] == "a single value":
                            file_storage = form[key].data
                            filename = file_storage.filename
                            if filename in files:
                                c = 2
                                tmp_name = f"{c}__{filename}"
                                while tmp_name in files:
                                    c += 1
                                    tmp_name = f"{c}__{filename}"
                                filename = tmp_name
                            files[filename] = file_storage
                            required.append(f"job:data/{filename}")
                        else:
                            for _file in form[key].data:
                                file_storage = _file
                                filename = file_storage.filename
                                if filename == "":
                                    continue
                                if filename in files:
                                    c = 2
                                    tmp_name = f"{c}__{filename}"
                                    while tmp_name in files:
                                        c += 1
                                        tmp_name = f"{c}__{filename}"
                                    filename = tmp_name
                                files[filename] = file_storage
                                required.append(f"job:data/{filename}")
                    else:
                        control_parameters[name] = result[key]
                        if values["nargs"] == "a single value":
                            required.append(result[key])
                        else:
                            required.extend(shlex.split(result[key]))

            if len(required) > 0:
                optional.append("--")
                cmdline = optional + required
            else:
                cmdline = optional

        # Submit the job ... how do we do this?
        job = util.setup_job(
            flowchart_text,
            result["project"],
            result["title"],
            result["description"],
            {"cmdline": cmdline, "control parameters": control_parameters},
        )
        job_id = job.id

        # Save any files transferred with the job
        if len(files) > 0:
            path = Path(job.path) / "data"
            path.mkdir(exist_ok=True)
            for filename, file_storage in files.items():
                file_storage.save(path / filename)
                file_storage.close()

        # Let the user know the job was actually submitted.
        flash(f"Submitted as job {job_id}.")

        # If just submitting one job go to its page
        if result["submit"]:
            base_url = url_for("main.index")
            job_url = base_url + f"#jobs/{job_id}"
            return redirect(job_url)

    return render_template("jobs/job_parameters.html", form=form)


@jobs.route("/jobs/<id>/resubmit", methods=["GET", "POST"])
@jwt_required()
def resubmit_job(id=None):
    """Handle the dialogs for submitting a job by uploading a flowchart.

    This is step 2: setting up the page to get the parameters
    
    Steps:
        1. Get the file
        2. Read the file to setup the dialog with the correct parameters
    """
    global flowchart_text

    # Get the flowchart   TODO permissions???
    job = Job.get_by_id(id)

    # The flowchart
    path = Path(job.path) / "flowchart.flow"
    try:
        text = path.read_text()
    except Exception:
        flash("An error occured reading the flowchart!")
        base_url = url_for("main.index")
        job_url = base_url + f"#jobs/{job.id}"
        return redirect(job_url)

    flowchart_text = text

    return job_parameters(template_job=job)

@jobs.after_request
def job_refresh(response):
    return refresh_expiring_jwts(response)
